@@ -3,13 +3,16 @@ import time
 import json
 import logging
 import signal
+import pickle
 import subprocess
 import pandas as pd
 import datetime
-from os.path import join, split, exists
+from os.path import join, split
 from d3m_interface.basic_ta3 import BasicTA3
 from d3m_interface.visualization import plot_metadata, plot_comparison_pipelines, plot_text_summary
 from d3m_interface.data_converter import is_d3m_format, dataset_to_d3m, d3mtext_to_dataframe, copy_folder, to_d3m_json
+from d3m_interface.pipeline import Pipeline
+from d3m_interface.confidence_calculator import create_confidence_pipeline
 from d3m.metadata.problem import PerformanceMetric
 from threading import Thread
 from IPython.core.getipython import get_ipython
@@ -17,10 +20,10 @@ from IPython.core.getipython import get_ipython
 
 logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s', stream=sys.stdout)
 logger = logging.getLogger(__name__)
-pd.set_option('display.max_colwidth', None)
+
 
 TA2_DOCKER_IMAGES = {'AlphaD3M': 'registry.gitlab.com/vida-nyu/d3m/alphad3m:latest',
-                     'CMU': 'registry.datadrivendiscovery.org/sheath/cmu-ta2:latest',
+                     'CMU': 'registry.gitlab.com/sray/cmu-ta2:latest',
                      'SRI': 'registry.gitlab.com/daraghhartnett/autoflow:latest',
                      'TAMU': 'dmartinez05/tamuta2:latest'}
 
@@ -45,11 +48,11 @@ class AutoML:
         """Create/instantiate an AutoML object
 
         :param output_folder: Path to the output directory
-        :param ta2_id: TA2 system name. It makes reference to the TA2 docker image. The provided TA2 systems are the
-            following: `AlphaD3M, CMU, SRI, TAMU`
+        :param ta2_id: AutoML system name. It makes reference to the AutoML docker image. The provided AutoML systems
+            are the following: `AlphaD3M, CMU, SRI, TAMU`
         """
         if ta2_id not in TA2_DOCKER_IMAGES:
-            raise ValueError('Unknown "%s" TA2, you should choose among: [%s]' % (ta2_id, ', '.join(TA2_DOCKER_IMAGES)))
+            raise ValueError('Unknown "%s" AutoML, you should choose among: [%s]' % (ta2_id, ', '.join(TA2_DOCKER_IMAGES)))
 
         self.output_folder = output_folder
         self.ta2_id = ta2_id
@@ -89,7 +92,9 @@ class AutoML:
         :returns: List of pipelines
         """
         suffix = 'TRAIN'
+        dataset_in_container = '/input/dataset/'
         signal.signal(signal.SIGINT, kernel_interrupt_handler)
+
         if not is_d3m_format(dataset, suffix):
             self.problem_config = {'target_column': target, 'metric': metric, 'task_keywords': task_keywords,
                                    'optional': kwargs}
@@ -100,11 +105,10 @@ class AutoML:
         search_id = None
         signal.signal(signal.SIGALRM, lambda signum, frame: self.ta3.do_stop_search(search_id))
         signal.alarm(time_bound * 60)
-
-        dataset_in_container = '/input/dataset/TRAIN/dataset_TRAIN/datasetDoc.json'
+        train_dataset_d3m = join(dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
         problem_path = join(dataset, 'problem_TRAIN/problemDoc.json')
         start_time = datetime.datetime.utcnow()
-        pipelines = self.ta3.do_search(dataset_in_container, problem_path, time_bound, time_bound_run)
+        pipelines = self.ta3.do_search(train_dataset_d3m, problem_path, time_bound, time_bound_run)
 
         jobs = []
 
@@ -122,7 +126,7 @@ class AutoML:
             duration = str(end_time - start_time)
             logger.info('Found pipeline id=%s, time=%s, scoring...' % (pipeline['id'], duration))
 
-            job = Thread(target=self.score_in_search, args=(pipeline, dataset_in_container, problem_path, self.pipelines,
+            job = Thread(target=self.score_in_search, args=(pipeline, train_dataset_d3m, problem_path, self.pipelines,
                                                             method, stratified, shuffle, folds, train_ratio, random_seed))
             jobs.append(job)
             job.start()
@@ -147,6 +151,7 @@ class AutoML:
             self.leaderboard = pd.DataFrame(leaderboard, columns=['ranking', 'id', 'summary', metric])
 
         signal.alarm(0)
+
         return self.pipelines.values()
 
     def train(self, pipeline_id, expose_outputs=None):
@@ -158,8 +163,7 @@ class AutoML:
             ids of the steps, e.g. 'steps.2.produce'
         :returns: A dictionary that contains the id and fitted model with/without the pipeline step outputs
         """
-
-        dataset_in_container = '/input/dataset/TRAIN/dataset_TRAIN/datasetDoc.json'
+        dataset_in_container = '/input/dataset/'
 
         if pipeline_id not in self.pipelines:
             raise ValueError('Pipeline id=%s does not exist' % pipeline_id)
@@ -173,8 +177,17 @@ class AutoML:
                 for id_output in step['outputs']:
                     expose_outputs.append('steps.%d.%s' % (index, id_output['id']))
 
-        fitted_pipeline_id, pipeline_step_outputs = self.ta3.do_train(pipeline_id, dataset_in_container, expose_outputs)
-        fitted_pipeline = None  # TODO: Call to LoadFittedSolution, but TA2 could not have implemented it yet
+        train_dataset_d3m = join(dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
+        fitted_pipeline_id, pipeline_step_outputs = self.ta3.do_train(pipeline_id, train_dataset_d3m, expose_outputs)
+        fitted_pipeline = None
+        # TODO: Implement other method to save the fitted pipeline
+        # try:
+        #     fitted_solution_uri = self.ta3.do_save_fitted_solution(fitted_pipeline_id)
+        #     fitted_solution_uri = fitted_solution_uri.replace('file:///output/', '')
+        #     with open(join(self.output_folder, fitted_solution_uri), 'rb') as fin:
+        #         fitted_pipeline = pickle.load(fin)
+        # except Exception as e:
+        #     logger.warning('Fitted pipeline id=%s could not be loaded' % pipeline_id, exc_info=e)
 
         for step_id, step_csv_uri in pipeline_step_outputs.items():
             if not step_csv_uri.startswith('file://'):
@@ -193,7 +206,7 @@ class AutoML:
 
         return model, pipeline_step_outputs
 
-    def test(self, model, test_dataset, expose_outputs=None):
+    def test(self, model, test_dataset, expose_outputs=None, calculate_confidence=False):
         """Test a model
 
         :param model: Dict that contains the id and fitted model
@@ -201,21 +214,64 @@ class AutoML:
         :param expose_outputs: The output of the pipeline steps. If None, it doesn't expose any output of the steps.
             If str, should be 'all' to shows the output of each step in the pipeline, If list, it should contain the
             ids of the steps, e.g. 'steps.2.produce'
+        :param calculate_confidence: Whether or not to return the confidence instead of the predictions
         :returns: A dataframe that contains the predictions with/without the pipeline step outputs
         """
         suffix = 'TEST'
-        dataset_in_container = '/input/dataset/TEST/dataset_TEST/datasetDoc.json'
+        dataset_in_container = '/input/dataset/'
 
         if not is_d3m_format(test_dataset, suffix):
             dataset_to_d3m(test_dataset, self.output_folder, self.problem_config, suffix)
         elif test_dataset != join(self.dataset, 'TEST'):  # Special case for D3M test dataset with different path
             destination_path = join(self.output_folder, 'temp', 'dataset_d3mformat', 'TEST')
             copy_folder(test_dataset, destination_path)
-            dataset_in_container = '/output/temp/dataset_d3mformat/TEST/dataset_TEST/datasetDoc.json'
+            dataset_in_container = '/output/temp/dataset_d3mformat/'
 
-        pipeline_id = model['id']
-        fitted_pipeline_id = self.pipelines[pipeline_id]['fitted_id']
         logger.info('Testing model...')
+        pipeline_id = model['id']
+        test_dataset_d3m = join(dataset_in_container, 'TEST/dataset_TEST/datasetDoc.json')
+
+        if calculate_confidence:
+            # The only way to get the confidence is through the CLI utility, TA3TA2 API doesn't support it
+            original_pipeline = self.pipelines[pipeline_id]['json_representation']
+            confidence_pipeline = create_confidence_pipeline(original_pipeline)
+
+            with open(join(self.output_folder, '%s.json' % pipeline_id), 'w') as fout:
+                json.dump(confidence_pipeline, fout)  # Save temporally the json pipeline
+
+            train_dataset_d3m = join(dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
+            problem_path = join(dataset_in_container, 'TRAIN/problem_TRAIN/problemDoc.json')
+            pipeline_path = join('/output/', '%s.json' % pipeline_id)
+            output_csv_path = join('/output/', 'fit_produce_%s.csv' % pipeline_id)
+
+            process = subprocess.Popen(
+                [
+                    'docker', 'exec', 'ta2_container',
+                    'python3', '-m', 'd3m',
+                    'runtime',
+                    '--context', 'TESTING',
+                    '--random-seed', '0',
+                    'fit-produce',
+                    '--pipeline', pipeline_path,
+                    '--problem', problem_path,
+                    '--input', train_dataset_d3m,
+                    '--test-input', test_dataset_d3m,
+                    '--output', output_csv_path,
+                ],
+                stderr=subprocess.PIPE
+            )
+            _, stderr = process.communicate()
+
+            if process.returncode != 0:
+                raise RuntimeError(stderr.decode())
+
+            result_path = join(self.output_folder, 'fit_produce_%s.csv' % pipeline_id)
+            predictions = pd.read_csv(result_path)
+            logger.info('Testing finished!')
+
+            return predictions
+
+        fitted_pipeline_id = self.pipelines[pipeline_id]['fitted_id']
 
         if expose_outputs is None:
             expose_outputs = []
@@ -229,7 +285,7 @@ class AutoML:
         if 'outputs.0' not in expose_outputs:
             expose_outputs.append('outputs.0')
 
-        pipeline_step_outputs = self.ta3.do_test(fitted_pipeline_id, dataset_in_container, expose_outputs)
+        pipeline_step_outputs = self.ta3.do_test(fitted_pipeline_id, test_dataset_d3m, expose_outputs)
 
         for step_id, step_csv_uri in pipeline_step_outputs.items():
             if not step_csv_uri.startswith('file://'):
@@ -255,28 +311,28 @@ class AutoML:
         :returns: A tuple holding metric name and score value
         """
         suffix = 'SCORE'
+        dataset_in_container = '/input/dataset/'
 
         if not is_d3m_format(test_dataset, suffix):
             dataset_to_d3m(test_dataset, self.output_folder, self.problem_config, suffix)
 
-        if not isinstance(pipeline_id, str):
-            d3m_json = to_d3m_json(pipeline_id)
-            pipeline_id = d3m_json['id']
-        else:
+        if isinstance(pipeline_id, Pipeline):
+            pipeline_json = to_d3m_json(pipeline_id)
+            pipeline_id = pipeline_json['id']
+        elif isinstance(pipeline_id, str):
             if pipeline_id not in self.pipelines:
                 raise ValueError('Pipeline id=%s does not exist' % pipeline_id)
-            d3m_json = self.pipelines[pipeline_id]['json_representation']
+            pipeline_json = self.pipelines[pipeline_id]['json_representation']
 
         with open(join(self.output_folder, '%s.json' % pipeline_id), 'w') as fout:
-            json.dump(d3m_json, fout)  # Save temporally the json pipeline
+            json.dump(pipeline_json, fout)  # Save temporally the json pipeline
 
-        dataset_in_container = '/input/dataset/'
-        dataset_train_path = join(dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
-        dataset_test_path = join(dataset_in_container, 'TEST/dataset_TEST/datasetDoc.json')
-        dataset_score_path = join(dataset_in_container, 'SCORE/dataset_SCORE/datasetDoc.json')
+        train_dataset_d3m = join(dataset_in_container, 'TRAIN/dataset_TRAIN/datasetDoc.json')
+        test_dataset_d3m = join(dataset_in_container, 'TEST/dataset_TEST/datasetDoc.json')
+        score_dataset_d3m = join(dataset_in_container, 'SCORE/dataset_SCORE/datasetDoc.json')
         problem_path = join(dataset_in_container, 'TRAIN/problem_TRAIN/problemDoc.json')
         pipeline_path = join('/output/', '%s.json' % pipeline_id)
-        score_pipeline_path = join('/output/', 'fit_score_%s.csv' % pipeline_id)
+        output_csv_path = join('/output/', 'fit_score_%s.csv' % pipeline_id)
 
         #  TODO: Use TA2TA3 API to score
         process = subprocess.Popen(
@@ -289,10 +345,10 @@ class AutoML:
                 'fit-score',
                 '--pipeline', pipeline_path,
                 '--problem', problem_path,
-                '--input', dataset_train_path,
-                '--test-input', dataset_test_path,
-                '--score-input', dataset_score_path,
-                '--scores', score_pipeline_path
+                '--input', train_dataset_d3m,
+                '--test-input', test_dataset_d3m,
+                '--score-input', score_dataset_d3m,
+                '--scores', output_csv_path
             ],
             stderr=subprocess.PIPE
         )
@@ -371,7 +427,7 @@ class AutoML:
         return profiler_inputs
 
     def export_pipeline_code(self, pipeline_id, ipython_cell=True):
-        """Converts a Pipeline Description to an executable python script
+        """Converts a Pipeline Description to an executable Python script
 
         :param pipeline_id: Pipeline id
         :param ipython_cell: Whether or not to show the Python code in a Jupyter Notebook cell
@@ -478,12 +534,12 @@ class AutoML:
 
                 time.sleep(4)
 
-    def score_in_search(self, pipeline, dataset_in_container, problem_path, pipelines, method, stratified, shuffle,
-                        folds, train_ratio, random_seed):
+    def score_in_search(self, pipeline, train_dataset, problem_path, pipelines, method, stratified, shuffle, folds,
+                        train_ratio, random_seed):
         try:
             pipeline['start_time'] = datetime.datetime.utcnow().isoformat() + 'Z'
-            score_data = self.ta3.do_score(pipeline['id'], dataset_in_container, problem_path, method, stratified,
-                                           shuffle, folds, train_ratio, random_seed)
+            score_data = self.ta3.do_score(pipeline['id'], train_dataset, problem_path, method, stratified, shuffle,
+                                           folds, train_ratio, random_seed)
             logger.info('Scored pipeline id=%s, %s=%s' % (pipeline['id'], score_data['metric'], score_data['score']))
             pipeline['end_time'] = datetime.datetime.utcnow().isoformat() + 'Z'
             pipeline['score'] = score_data['score']
@@ -504,20 +560,25 @@ class AutoML:
 
         return ', '.join(primitives_summary)
 
-    def plot_summary_dataset(self, dataset_path):
-        """ Plot histograms of the dataset
-
-        :param dataset_path: Path to dataset.  It supports D3M dataset, and CSV file
+    def plot_leaderboard(self):
+        """Plot pipelines' leaderboard
         """
-        suffix = dataset_path.split('/')[-1]
+        return self.leaderboard.style.hide_index()
 
-        if is_d3m_format(dataset_path, suffix):
-            dataset_path = join(dataset_path, 'dataset_%s/tables/learningData.csv' % suffix)
+    def plot_summary_dataset(self, dataset):
+        """Plot histograms of the dataset
 
-        plot_metadata(dataset_path)
+        :param dataset: Path to dataset.  It supports D3M dataset, and CSV file
+        """
+        suffix = dataset.split('/')[-1]
+
+        if is_d3m_format(dataset, suffix):
+            dataset = join(dataset, 'dataset_%s/tables/learningData.csv' % suffix)
+
+        plot_metadata(dataset)
 
     def plot_comparison_pipelines(self, test_dataset=None, source_name=None):
-        """ Plot PipelineProfiler visualization
+        """Plot PipelineProfiler visualization
 
         :param test_dataset: Path to dataset. If None it will use the search scores, otherwise will score the
             pipelines over the passed dataset
@@ -526,14 +587,14 @@ class AutoML:
         pipelineprofiler_inputs = self.create_pipelineprofiler_inputs(test_dataset, source_name)
         plot_comparison_pipelines(pipelineprofiler_inputs)
 
-    def plot_text_analysis(self, dataset_path, label_column, text_column):
-        """ Plot a visualization for text datasets
+    def plot_text_analysis(self, dataset, label_column, text_column):
+        """Plot a visualization for text datasets
 
-        :param dataset_path: Path to dataset.  It supports D3M dataset
+        :param dataset: Path to dataset.  It supports D3M dataset
         :param label_column: Name of the column that contains the categories
         :param text_column: Name of the column that contains the texts
         """
-        dataframe = d3mtext_to_dataframe(dataset_path, text_column)
+        dataframe = d3mtext_to_dataframe(dataset, text_column)
         plot_text_summary(dataframe, text_column, label_column)
 
     @staticmethod
